﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Net;
using System.Net.Sockets;

using System.Threading;
using System.IO;
namespace YSocket.AsyncSocket
{
    public class YunAsyncServerSocket : YunAsyncCommon
    {
        Task _vk = null;
        List<Socket> _sockets = new List<Socket>();
        Dictionary<Socket, CancellationTokenSource> _dic = new Dictionary<Socket, CancellationTokenSource>();
        public override void MkNm()
        {
            Nm= $"서버[{LocalSocketIpEndPoint.Address}:{LocalSocketIpEndPoint.Port}]";
        }

        public async Task<bool>  Start()
        {
            if (!CheckSocketStart())
            {
                return false;
            }
            LocalSocket = GetTcpSocket();
            LocalSocket.Bind(LocalSocketIpEndPoint);
            LocalSocket.Listen(5);

            ActSysMessage?.Invoke("서버시작");
            LoopAccept();
            return true;
        }

        /// <summary>
        /// 클라이언트의 접속을 기다리고 접속이 처리되면 반복적으로 처리
        /// </summary>
        /// <returns></returns>
        private void LoopAccept()
        {
            var awaiter = AcceptTaskCompletionSource(LocalSocket).GetAwaiter();
            awaiter.OnCompleted(() => {
                var client = awaiter.GetResult();
                _sockets.Add(client);
                LoopReceive(client);
                LoopAccept();
            });
        }

        

        private  void Isworking()
        {
            Func<Socket, bool> func;
            func = (a) =>
            {
                try
                {
                    int recv = a.Send(new byte[1]);
                    if (recv > 0) { return true; }
                    return false;
                }
                catch (Exception er)
                {

                    return false;
                }
            };
            var t =Task.Run( async() => {

                await Task.Delay(2000);
                GC.Collect();
                var deleteList= _sockets.Where(s => !func(s));
                foreach (var item in deleteList)
                {
                    _sockets.Remove(item);
                    item.Close();
                }
                
            });
            var awaiter= t.GetAwaiter();
            awaiter.OnCompleted(() => {
                Isworking();
            });

        }


        /// <summary>
        /// 소켓이 살아있는 동안 데이터를 받고 완료후에  다시 데이터를 받는 동작을 처리함.
        /// </summary>
        /// <param name="socket"></param>
        private void LoopReceive( Socket socket)
        {
            if (socket == null || !socket.Connected)
            {
                return;
            }
            var awaiter = ReceiveTaskCompletionSource(socket).GetAwaiter();
            awaiter.OnCompleted(() => {
                var ret = awaiter.GetResult();
                if (ret ==null || ret.Length == 0)
                {
                    return;
                }
                string recv = Encoding.UTF8.GetString(ret);
                ActReceiveMessage?.Invoke(recv);
                LoopReceive(socket);
            });
        }

        private async void LoopReceive(Socket socket ,CancellationToken token)
        {
            if (socket == null || !socket.Connected)
            {
                return;
            }
            var ret = await ReceiveTaskCompletionSource(socket ,token);
            if (ret == null)
            {
                return;
            }
            else if (ret.Length < 1)
            {
                return;
            }
            else
            {
                string recv = Encoding.UTF8.GetString(ret);
                ActReceiveMessage?.Invoke(recv);
                LoopReceive(socket ,token);
            }
        }

        private void 자료받기(IAsyncResult ar)
        {
            var rsocket=  ar.AsyncState as Socket;
            int rbyte= rsocket.EndReceive(ar);
        }



        /// <summary>
        /// beginReceive가 완료되면 endReceive로 완료처리를 한후에  TaskCompletionSource를 통해서
        /// 받은 데이터를 전송함.
        /// </summary>
        /// <param name="socket"></param>
        /// <returns></returns>
        public  Task<Byte[]> ReceiveTaskCompletionSource (Socket socket)
        {

            if (socket == null || !socket.Connected)
            {
                return null;
            }
            var tcs = new TaskCompletionSource<Byte[]>(socket);
            var buffer = new byte[1024];
   
            socket.BeginReceive(buffer, 0, 1024, SocketFlags.None, ar => {
                try
                {
                    var ret = ar.AsyncState as Socket;
                    int retcnt = ret.EndReceive(ar);
                    var retbuffer = new byte[retcnt];
                    Buffer.BlockCopy(buffer, 0, retbuffer, 0, retcnt);
                    tcs.SetResult(retbuffer);
                }
                catch (Exception er)
                {
                    tcs.SetException(er);
                }
            }, socket);
   
            return tcs.Task;
        }

        public async Task<Byte[]> ReceiveTaskCompletionSource(Socket socket ,CancellationToken token)
        {

            if (socket == null || !socket.Connected)
            {
                return null;
            }
            var tcs = new TaskCompletionSource<Byte[]>(socket);
            var buffer = new byte[1024];


            socket.BeginReceive(buffer, 0, 1024, SocketFlags.None, ar => {
                try
                {
                    var ret = ar.AsyncState as Socket;
                    int retcnt = ret.EndReceive(ar);
                    var retbuffer = new byte[retcnt];
                    Buffer.BlockCopy(buffer, 0, retbuffer, 0, retcnt);
                   
                    tcs.SetResult(retbuffer);

                }
                catch (Exception er)
                {
                    tcs.SetException(er);
                }
            }, socket);
            if (token.IsCancellationRequested)
            {
                tcs.SetCanceled();
            }
            return  tcs.Task.Result;
        }
        /// <summary>
        /// Socket의 beginAccept가 완료되면 endAccept로 완료처리를 하고 client를 
        /// TaskCompletionSource를 통해서 전달함 block을 사용하지 않는게 목표.
        /// </summary>
        /// <param name="socket"></param>
        /// <returns></returns>
        public Task<Socket> AcceptTaskCompletionSource (Socket socket)
        {
            var tcs = new TaskCompletionSource<Socket>(socket);
            socket.BeginAccept(a => {
                try
                {
                    var retsocket= a.AsyncState as Socket;
                    var client = retsocket.EndAccept(a);
                    tcs.SetResult(client);
                }
                catch (Exception e)
                {
                    tcs.SetException(e);
                }

            }, socket);
            return tcs.Task;
        }

        private void 접속완료(object sender, SocketAsyncEventArgs accepteEar)
        {
            ActSysMessage?.Invoke("접속완료");
            var accept= accepteEar.AcceptSocket;
            
            SocketAsyncEventArgs receiveEar = new SocketAsyncEventArgs();

            receiveEar.SetBuffer(0,1024);
    
            receiveEar.Completed += 수신완료;
            accepteEar.AcceptSocket = null;
            try
            {
            accept.ReceiveAsync(receiveEar);    

            }
            catch (Exception er)
            {

                ActSysMessage?.Invoke(er.Message);
            }
            accepteEar.AcceptSocket = null;
            LocalSocket.AcceptAsync(accepteEar);
        }

        internal void SendMessage(string text)
        {
            throw new NotImplementedException();
        }

        private void 수신완료(object sender, SocketAsyncEventArgs e)
        {
            ActSysMessage?.Invoke("수신완료");
            var r = e.BufferList;
            var r2 = e.Buffer;
            var r3 = e.BytesTransferred;
               
        }

        internal void CloseWaitingThread(int stopthid)
        {
            throw new NotImplementedException();
        }



        internal void Stop()
        {
            throw new NotImplementedException();
        }
    }
}
